package main

import (
	"bufio"
	"code.google.com/p/go.crypto/blowfish"
	"encoding/gob"
	"flag"
	"fmt"
	"log"
	"os"
)

func usage() {
	fmt.Fprintf(os.Stderr, "usage: %s <pass> <encrypted-file>\n",
		os.Args[0])
	os.Exit(2)
}

func decryptBuf(c *blowfish.Cipher, a []byte) []byte {
	pt := make([]byte, len(a))
	for i := 0; i < len(a); i += blowfish.BlockSize {
		c.Decrypt(pt[i:i+blowfish.BlockSize],
			a[i:i+blowfish.BlockSize])
	}
	return pt
}

type hdr struct {
	Magic string
	Len   int
	Data  []byte
}

func main() {
	log.SetPrefix("bdec: ")
	flag.Usage = usage
	flag.Parse()

	args := flag.Args()
	if len(args) < 2 {
		usage()
	}

	c, err := blowfish.NewCipher([]byte(args[0]))
	if err != nil {
		log.Fatal(err)
	}

	fi, err := os.Open(args[1])
	if err != nil {
		log.Fatal(err)
	}
	defer fi.Close()

	r := bufio.NewReader(fi)
	g := gob.NewDecoder(r)

	h := new(hdr)
	err = g.Decode(h)
	if err != nil {
		log.Fatal(err)
	}

	if h.Magic != "BENC" {
		log.Fatal(fmt.Errorf("Invalid magic: %s", h.Magic).Error())
	}

	db := decryptBuf(c, h.Data)
	fmt.Printf("%s", db[0:h.Len])
}
