package main

import (
	"code.google.com/p/go.crypto/blowfish"
	"encoding/gob"
	"flag"
	"fmt"
	"io/ioutil"
	"log"
	"os"
)

func usage() {
	fmt.Fprintf(os.Stderr, "usage: %s <pass> <file>\n",
		os.Args[0])
	os.Exit(2)
}

func encryptBuf(c *blowfish.Cipher, a []byte) []byte {
	padded := a
	diff := len(a) % blowfish.BlockSize
	if diff != 0 {
		diff = blowfish.BlockSize - diff
	}
	for i := 0; i < diff; i++ {
		padded = append(padded, byte(0x0))
	}
	ct := make([]byte, len(padded))
	for i := 0; i < len(a); i += blowfish.BlockSize {
		if i+blowfish.BlockSize > len(a) {
			c.Encrypt(ct[i:], padded[i:])
			return ct
		}
		c.Encrypt(ct[i:i+blowfish.BlockSize],
			padded[i:i+blowfish.BlockSize])
	}
	return ct
}

type hdr struct {
	Magic string
	Len   int
	Data  []byte
}

func main() {
	log.SetPrefix("benc: ")
	flag.Usage = usage
	flag.Parse()

	args := flag.Args()
	if len(args) < 2 {
		usage()
	}

	c, err := blowfish.NewCipher([]byte(args[0]))
	if err != nil {
		log.Fatal(err)
	}

	b, err := ioutil.ReadFile(args[1])
	if err != nil {
		log.Fatal(err)
	}
	eb := encryptBuf(c, b)

	h := &hdr{"BENC", len(b), eb}
	g := gob.NewEncoder(os.Stdout)
	err = g.Encode(*h)
	if err != nil {
		log.Fatal(err)
	}
}
