// Copyright 2013 TLH and dsp. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package games

import (
	"irc"
)

// This is a base gamestate for irc games
type gameState struct {
	ictx    *irc.Context      // The context to send commands with
	gchan   *chan irc.Message // A chan that intercepts game irc messages
	name    string            // Name of the game
	creator string            // Name of game creator
}

// A GameEnger will List available games and invoke a new one if it is on the map.
// New games are invoked as new goroutines that communicate via channels.
type GameEnger interface {
	List() []string
	New(string, *irc.Context, *chan irc.Message)
}

type gameEng struct {
	games map[string]func(*irc.Context, *chan irc.Message)
}

// Returns all available games as a string array
func (g *gameEng) List() []string {
	r := []string{}
	for k, _ := range g.games {
		r = append(r, k)
	}
	return r
}

// Initializes games strings with their corresponding goroutines
func NewGameEng() *gameEng {
	return &gameEng{games: map[string]func(*irc.Context, *chan irc.Message){"hangman": cmdHangman}}

}

// Fire up a goroutine if the game is in the map
func (g *gameEng) New(a string, ic *irc.Context, gc *chan irc.Message) {
	if f, ok := g.games[a]; ok {
		go f(ic, gc)
	}
}
