// Copyright 2013 TLH and dsp. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package irc

import (
	"fmt"
)

type Event struct {
	Command string
	Fn      func(Message)
}

func (i *Context) AddEventHandler(ev Event) error {
	i.Lock()
	defer i.Unlock()
	for j := 0; j < i.evnum; j++ {
		if i.ev[j].Command == ev.Command {
			return fmt.Errorf("Event %s already handled",
				ev.Command)
		}
	}
	i.ev = append(i.ev, ev)
	i.evnum++
	return nil
}

func (i *Context) DelEventHandler(ev *Event) error {
	i.Lock()
	defer i.Unlock()
	for j := 0; j < i.evnum; j++ {
		if &i.ev[j] == ev {
			i.ev = append(i.ev[:j], i.ev[j+1:]...)
			return nil
		}
	}
	return fmt.Errorf("Event handler not found")
}

func (i *Context) dispatchHandler(msg Message) {
	i.Lock()
	defer i.Unlock()
	for _, v := range i.ev {
		if v.Command == msg.Command {
			v.Fn(msg)
			break
		}
	}
}
