// Copyright 2013 TLH and dsp. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package irc

import (
	"crypto/tls"
	"unicode"
)

func loadCerts() (*tls.Config, error) {
	cert, err := tls.LoadX509KeyPair("certs/client.pem",
		"certs/client.key")
	if err != nil {
		return nil, err
	}
	return &tls.Config{
		Certificates:       []tls.Certificate{cert},
		InsecureSkipVerify: true,
	}, nil
}

func validNick(n string) bool {
	for i, _ := range n {
		c := n[i]
		if !unicode.IsLetter(rune(c)) &&
			!unicode.IsDigit(rune(c)) {
			switch c {
			case '[', ']', '\\', '`', '_', '^', '{', '|', '}', '-':
			default:
				return false
			}
		}
	}
	return true
}
