// Copyright 2013 TLH and dsp. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package main

import (
	"flag"
	"fmt"
	"games"
	"io"
	"io/ioutil"
	"irc"
	"log"
	"mapfs"
	"math/rand"
	"net/http"
	"os"
	"os/exec"
	"regexp"
	"strings"
	"syscall"
	"time"
)

type kuntCtx struct {
	stime  time.Time
	ircCtx *irc.Context
}

func cmdKunt(msg irc.Message) {
	replies := []string{
		"dn niotho maaaan",
		"eimai kokalo telios... gama ta",
		"pwpw gamise ta, den vlepo brosta mou",
		"yoyoyoyo duuuuude",
		"die",
		"ok",
		"maybe",
		"yes",
	}
	if msg.Nick == "dsp" || msg.Nick == "TLH" || msg.Nick == "sovarofanis" {
		replies = []string{
			"You are my Mastah!",
			"I will always obey you my Mastah!",
			"vv maaaaan",
			"etimazo na stripso ena ro... ade vives",
			"420 dawwwwwwg",
			"You own all ma base niggah, chiiilll",
			"spliff it up bro!",
		}
	}
	if strings.Index(msg.Args[1], "vv") != -1 ||
		strings.Index(msg.Args[1], "vives") != -1 {
		replies = []string{
			"vives alani",
			"vv maaaan!!",
			"skastoooo",
			"vv http://www.youtube.com/watch?v=wHzxh9fneJ8 <-- varaeeiiiiii",
			"vives!",
			"vv!",
			"vv dawg",
		}
	}
	if strings.Index(msg.Args[1], "vi ") != -1 ||
		strings.Index(msg.Args[1], " vi") != -1 ||
		strings.Index(msg.Args[1], "vim ") != -1 ||
		strings.Index(msg.Args[1], " vim") != -1 {
		replies = []string{
			"emax ftw",
			"emacs ftw!",
			"EMACS!",
			"ed is the standard editor.",
			"emax or die",
			"mg or die",
			"vi sux",
			"vim sux ass",
		}
	}
	idx := rand.Intn(len(replies))
	kunt.ircCtx.Privmsg(msg.Args[0], string(replies[idx]))
}

func cmdWisdom(msg irc.Message) {
	out, err := exec.Command("./fortune").Output()
	if err != nil {
		log.Fatal(err)
	}
	lines := strings.Split(string(out), "\n")
	for _, i := range lines {
		if i == "" {
			continue
		}
		kunt.ircCtx.Privmsg(msg.Args[0], i)
		time.Sleep(512 * time.Millisecond)
	}
}

func cmdHelp(msg irc.Message) {
	help := []string{
		"!addquote <quote> -> Add a quote to the db",
		"!randquote        -> Print a random quote from the db",
		"!countquotes      -> Print the number of entries in the quote db",
		"!addurl <url>     -> Add a url to the db",
		"!randurl          -> Print a random url from the db",
		"!counturls        -> Print the number of entries in the url db",
		"!uptime           -> Show uptime",
		"!wisdom           -> Invoke fortune(6)",
		"!radio            -> Print now playing song",
		"!game             -> List or Play games",
		"!src              -> Fetch kunt's source code",
		"!TODO             -> Print the TODO list",
		"!version          -> Show version",
		"!copyright        -> Show copyright info",
	}
	for _, i := range help {
		kunt.ircCtx.Privmsg(msg.Args[0], i)
		time.Sleep(512 * time.Millisecond)
	}
}

func cmdVersion(msg irc.Message) {
	ver := "v0.2.9"
	kunt.ircCtx.Privmsg(msg.Args[0], ver)
}

func cmdRandQuote(msg irc.Message) {
	if quoteDb.Empty() {
		kunt.ircCtx.Privmsg(msg.Args[0], "Empty quote database")
		return
	}
	quote, idx := quoteDb.Rand()
	quote = []byte(strings.TrimSpace(string(quote)))
	// Print idx in red
	text := fmt.Sprintf("%s%02d[%d]%s %s",
		"\003", '5', idx, "\003", string(quote))
	kunt.ircCtx.Privmsg(msg.Args[0], text)
}

func cmdAddQuote(msg irc.Message) {
	text := msg.Args[1]
	if len(strings.Fields(text)) < 2 {
		kunt.ircCtx.Privmsg(msg.Args[0], "Missing parameter for !addquote")
		return
	}
	text = text[10:]
	text = strings.TrimSpace(text)
	text += "\n"
	buf := []byte(text)
	if quoteDb.CountMatches(buf) > 0 {
		kunt.ircCtx.Privmsg(msg.Args[0],
			fmt.Errorf("Duplicate entry: %s", buf).Error())
		return
	}
	_, err := quoteDb.Append(buf)
	if err != nil {
		kunt.ircCtx.Privmsg(msg.Args[0], err.Error())
		return
	}
	err = quoteDb.Sync()
	if err != nil {
		log.Fatal(err)
	}
}

func cmdCountQuotes(msg irc.Message) {
	text := fmt.Sprintf("The quote DB has %d quotes", quoteDb.Size())
	kunt.ircCtx.Privmsg(msg.Args[0], text)
}

func cmdTheoQuote(msg irc.Message) {
	quotes := []string{
	"Write more code.",
	"Make more commits.",
	"That's because you have been slacking.",
	"slacker!",
	"That's what happens when you're lazy.",
	"idler!",
	"slackass!",
	"lazy bum!",
	"Stop slacking you lazy bum!",
	"slacker slacker lazy bum bum bum slacker!",
	"I could search... but I'm a lazy bum ;)",
	"sshutup sshithead, ssharpsshooting susshi sshplats ssharking assholes.",
	"Lazy bums slacking on your asses.",
	"35 commits an hour? That's pathetic!",
	"Fine software takes time to prepare.  Give a little slack.",
	"I am just stating a fact",
	"you bring new meaning to the terms slackass. I will have to invent a new term.",
	"if they cut you out, muddy their back yards",
	"Make them want to start over, and play nice the next time.",
	"It is clear that this has not been thought through.",
	"avoid using abort().  it is not nice.",
	"That's the most ridiculous thing I've heard in the last two or three minutes!",
	"I'm not just doing this for crowd response. I need to be right.",
	"I'd put a fan on my bomb.. And blinking lights...",
	"I love to fight",
	"No sane people allowed here.  Go home.",
	"you have to stop peeing on your breakfast",
	"feature requests come from idiots",
	"henning and darren / sitting in a tree / t o k i n g / a joint or three",
	"KICK ASS. TIME FOR A JASON LOVE IN!  WE CAN ALL GET LOST IN HIS HAIR!",
	"shame on you for following my rules.",
	"altq's parser sucks dead whale farts through the finest chemistry pipette's",
	"screw this operating system shit, i just want to drive!",
	"Search for fuck.  Anytime you see that word, you have a paragraph to write.",
	"Yes, but the ports people are into S&M.",
	"Buttons are for idiots.",
	"We are not hackers. We are turd polishing craftsmen.",
	"who cares.  style(9) can bite my ass",
	"It'd be one fucking happy planet if it wasn't for what's under this fucking sticker.",
	"I would explain, but I am too drunk.",
	"you slackers don't deserve pictures yet",
	"Vegetarian my ass",
	"Wait a minute, that's a McNally's!",
	"don't they recognize their moral responsibility to entertain me?",
	"#ifdef is for emacs developers.",
	"Many well known people become net-kooks in their later life, because they lose touch with reality.",
	"You're not allowed to have an opinion.",
	"tweep tweep tweep",
	"Quite frankly, SSE's alignment requirement is the most utterly retarded idea since eating your own shit.",
	"Holy verbose prom startup Batman.",
	"Any day now, when we sell out.",
	"optimism in man kind does not belong here",
	"First user who tries to push this button, he pounds into the ground with a rant of death.",
	"we did farts.  now we do sperm.  we are cutting edge.",
	"the default configuration is a mixture of piss, puke, shit, and bloody entrails.",
	"Stop wasting your time reading people's licenses.",
	"doing it with environment variables is OH SO SYSTEM FIVE LIKE OH MY GOD PASS ME THE SPOON",
	"Linux is fucking POO, not just bad, bad REALLY REALLY BAD",
	"penguins are not much more than chickens that swim.",
	"i am a packet sniffing fool, let me wipe my face with my own poo",
	"Whiners.  They scale really well.",
	"in your world, you would have a checklist of 50 fucking workarounds just to make a coffee.",
	"for once, I have nothing to say.",
	"You have no idea how fucked we are",
	"You can call it fart if you want to.",
	"wavelan is a battle field",
	"You are in a maze of gpio pins, all alike, all undocumented, and a few are wired to bombs.",
	"And that is why humppa sucks... cause it has no cause.",
	"cache aliasing is a problem that would have stopped in 1992 if someone had killed about 5 people who worked at Sun.",
	"Don't spread rumours about me being gentle.",
	"If municipal water filtering equipment was built by the gcc developers, the western world would be dead by now.",
	"kettenis supported a new machine in my basement and all I got to do was fix a 1 character typo in his html page commit.",
	"industry told us a lesson: when you're an asshole, they mail you hardware",
	"I was joking, really.  I think I am funny :-)",
	"the kernel is a harsh mistress",
	"Have I ever been subtle? If my approach ever becomes subtle, shoot me.",
	"the acpi stabs you in the back.  the acpi stabs you in the back. you die ...",
	"My cats are more observant than you.",
	"our kernels have no bugs",
	"style(9) has all these fascist rules, and i have a problem with some of them because i didn't come up with them",
	"I'm not very reliable",
	"I don't like control",
	"You aren't being conservative -- you are trying to be a caveman.",
	"nfs loves everyone",
	"basically, dung beetles fucking.  that's what kerberosV + openssl is like",
	"I would rather run Windows than use vi.",
	"if you assign that responsibility to non-hikers I will walk over and cripple you now.",
	"i ojbect two yoru splelng of achlhlocis.",
	"We have two kinds of developers - those that deal with their own shit and those that deal with other people's shit.",
	"If people keep adding such huge stuff, soon mg will be bigger than emacs.",
	"this change comes down to: This year, next year, 5 years from now, 10 years from now, or Oh fuck.",
	"backwards compatibility is king, and will remain king, until 2038.",
	"I don't know if the Internet's safe yet.",
	"Those who don't understand Unix are condemned to reinvent Multics in a browser",
	"Don't tell anybody I said that.",
	"Complaint forms are handled in another department.",
	"You'd be safer using Windows than the code which was just deleted.",
	"Shit should not be shared.",
	"the randomization in this entire codebase is a grand experiment in stupid",
	"My mailbox is full of shock.",
	"my integer overflow spidey senses are tingling.",
	"I'm just trying to improve the code...",
	"It's a pleasure to work on code you can't make worse.",
	"It's largely bad style to do (int)sizeof",
	"When I see Makefile.in, I know that \"in\" is short for \"insane\".",
	"This is the beer. And that's why we need a hackathon.",
	}

	idx := rand.Intn(len(quotes))
	kunt.ircCtx.Privmsg(msg.Args[0], quotes[idx])
}

func cmdRandUrl(msg irc.Message) {
	if urlDb.Empty() {
		kunt.ircCtx.Privmsg(msg.Args[0], "Empty url database")
		return
	}
	url, idx := urlDb.Rand()
	url = []byte(strings.TrimSpace(string(url)))
	// Print idx in red
	r := fmt.Sprintf("%s%02d[%d]%s %s",
		"\003", '5', idx, "\003", string(url))
	kunt.ircCtx.Privmsg(msg.Args[0], r)
}

func resolveYoutubeTitle(url string) (title string, ok bool) {
	title = ""
	ok = false
	if len(ytregex.FindAllString(url, -1)) > 0 {
		title = "Unresolved title"
		ok = true
		resp, err := http.Get(url)
		if err == nil {
			defer resp.Body.Close()
			contents, err := ioutil.ReadAll(resp.Body)
			if err == nil {
				titles := titleregex.FindAllString(string(contents), 1)
				if len(titles) > 0 {
					//this relates to the way titleregex is
					title = string(titles[0][35 : len(titles[0])-2])
				} else {
					log.Printf(" [resolveYoutubeTitle] could not find the title regexp for url: %s \n", url)
				}
			} else {
				log.Printf(" [resolveYoutubeTitle] could not read response body with ioutil\n")
			}
		} else {
			log.Printf(" [resolveYoutubeTitle] could not connect to youtube url for title fetching \n")
		}
		//print [Youtube] tag in purple
		title = fmt.Sprintf("        %s%02d[%s]%s %s",
			"\003", '6', "Youtube", "\003", string(title))
	}
	return
}

func cmdAddUrl(msg irc.Message) {
	text := msg.Args[1]
	title := ""
	if len(strings.Fields(text)) < 2 {
		kunt.ircCtx.Privmsg(msg.Args[0], "Missing parameter for !addurl")
		return
	}
	text = text[8:]
	text = strings.TrimSpace(text)
	// Check for yt link
	title, ok := resolveYoutubeTitle(text)
	if ok {
		text += title + "\n"
	} else {
		text += "\n"
	}
	buf := []byte(text)
	if urlDb.CountMatches(buf) > 0 {
		kunt.ircCtx.Privmsg(msg.Args[0],
			fmt.Errorf("Duplicate entry: %s", buf).Error())
		return
	}
	_, err := urlDb.Append(buf)
	if err != nil {
		kunt.ircCtx.Privmsg(msg.Args[0], err.Error())
		return
	}
	err = urlDb.Sync()
	if err != nil {
		log.Fatal(err)
	}
}

func cmdCountUrls(msg irc.Message) {
	r := fmt.Sprintf("The url DB has %d urls", urlDb.Size())
	kunt.ircCtx.Privmsg(msg.Args[0], r)
}

func cmdRadio(msg irc.Message) {
	np, e := http.Get("http://radio.2f30.org:8000/")
	if e != nil {
		kunt.ircCtx.Privmsg(msg.Args[0], "Is radio broken?")
	} else {
		defer np.Body.Close()
		info, _ := ioutil.ReadAll(np.Body)
		lines := strings.Split(string(info), "\n")
		for _, i := range lines {
			if i == "" {
				continue
			}
			kunt.ircCtx.Privmsg(msg.Args[0], i)
			time.Sleep(512 * time.Millisecond)
		}
	}
}

func cmdUptime(msg irc.Message) {
	etime := time.Now()
	r := fmt.Sprintf("%v", etime.Sub(kunt.stime))
	kunt.ircCtx.Privmsg(msg.Args[0], r)
}

func cmdSrc(msg irc.Message) {
	src := "http://git.2f30.org/kunt"
	kunt.ircCtx.Privmsg(msg.Args[0], src)
}

func cmdTodo(msg irc.Message) {
	todo, err := ioutil.ReadFile("TODO")
	if err != nil {
		kunt.ircCtx.Privmsg(msg.Args[0], "No TODO list available")
		return
	}
	sf := strings.FieldsFunc(string(todo), func(r rune) bool {
		switch r {
		case '\n':
			return true
		}
		return false
	})
	for _, v := range sf {
		kunt.ircCtx.Privmsg(msg.Args[0], v)
		time.Sleep(512 * time.Millisecond)
	}
}

func cmdGame(msg irc.Message) {
	text := msg.Args[1]
	if len(strings.Fields(text)) < 2 {
		for _, str := range geng.List() {
			kunt.ircCtx.Privmsg(msg.Args[0], "[game] "+str)
		}
		return
	}
	text = strings.TrimSpace(text[5:])
	found := false
	for _, str := range geng.List() {
		if str == text {
			found = true
			break
		}
	}
	if !found {
		kunt.ircCtx.Privmsg(msg.Args[0], "[game] game not found")
		return
	}
	gchan := make(chan irc.Message)
	geng.New(text, kunt.ircCtx, &gchan)
	time.Sleep(512 * time.Millisecond)
	gchan <- irc.Message{Command: "SETNICK", Args: []string{botname}} //Let the game engine know.
	gchan <- irc.Message{Command: "SETCHAN", Args: []string{msg.Args[0]}}
}

func cmdCopyright(msg irc.Message) {
	copyright := []string{
		"Copyright 2013 TLH and dsp. All rights reserved.",
		"Use of this source code is governed by a BSD-style",
		"license that can be found in the LICENSE file.",
	}
	for _, i := range copyright {
		kunt.ircCtx.Privmsg(msg.Args[0], i)
		time.Sleep(512 * time.Millisecond)
	}
}

func fifoCmdParse() {
	fi, err := os.Open(".kfifo")
	if err != nil {
		log.Fatal(err)
	}
	defer func() {
		if err := fi.Close(); err != nil {
			log.Fatal(err)
		}
	}()
	for {
		buf := make([]byte, 1024)
		n, err := fi.Read(buf)
		if err != nil && err != io.EOF {
			log.Fatal(err)
		}
		if n > 0 {
			buf = buf[0 : n-1]
			switch string(buf) {
			case "netjoin":
				kunt.ircCtx.JoinChannels()
			case "urlupdate":
				updateUrlDbTitles("db/urls-new")
			case "quit":
				donechan <- true
			default:
				log.Printf("Unknown fifo command:%s\n", string(buf))
			}
		}
	}
}

func updateUrlDbTitles(path string) {
	log.Printf("*** starting urlDb update under path %s ***\n", path)
	newurlPlain := mapfs.NewPlainMap()
	_, err := os.Stat(path)
	if err == nil {
		log.Printf(" [updateUrlDbTitles] for the db to be updated, please provide a non-existant directory to be populated. %s exists\n", path)
		return
	}
	newurlDb := mapfs.NewMapfs(newurlPlain, "Urls", path, "url")
	err = newurlDb.Load()
	if err != nil {
		log.Printf("error on loading new urlDb: %s\n", err.Error())
		return
	}
	iter := mapfs.MakeMapIter(urlDb)
	whitespacefunc := func(a rune) bool {
		return a == '\t' || a == ' '
	}
	for {
		_, v, ok := iter()
		if !ok {
			break
		}
		stringv := strings.TrimSpace(string(v))
		strs := strings.FieldsFunc(stringv, whitespacefunc)
		title, ok := resolveYoutubeTitle(strs[0])
		if ok {
			stringv = strs[0] + title + "\n"
		} else {
			stringv = strs[0] + "\n"
		}
		buf := []byte(stringv)
		_, err := newurlDb.Append(buf)
		if err != nil {
			log.Printf(" [updateUrlDbTitles] %s\n", err.Error())
			return
		}
		err = newurlDb.Sync()
		if err != nil {
			log.Fatal(err)
		}
	}
	log.Printf("*** urlDb update finished ***\n")
}


var quoteDb *mapfs.Mapfs
var urlDb *mapfs.Mapfs
var kunt kuntCtx
var geng games.GameEnger
var botname string
var ytregex *regexp.Regexp
var titleregex *regexp.Regexp
var configfile string
var sslon = flag.Bool("s", false, "SSL support")
var cfg *irc.Config
var donechan chan bool

func init() {
	flag.StringVar(&configfile,"c","","JSON configuration file")
}

func main() {
	kunt.stime = time.Now()
	botname = "kunt"
	geng = games.NewGameEng()
	log.SetPrefix("kunt: ")

	flag.Parse()
	if flag.NArg() < 1 && configfile == "" {
		fmt.Fprintf(os.Stderr, "usage: %s [-s] [-c configfile] host:port\n",
			os.Args[0])
		os.Exit(2)
	}


	quotePlain := mapfs.NewPlainMap()
	quoteDb = mapfs.NewMapfs(quotePlain, "Quotes", "db/quotes", "quote")
	urlPlain := mapfs.NewPlainMap()
	urlDb = mapfs.NewMapfs(urlPlain, "Urls", "db/urls", "url")

	err := quoteDb.Load()
	if err != nil {
		log.Fatal(err)
	}
	err = urlDb.Load()
	if err != nil {
		log.Fatal(err)
	}

	// Compile the yt regexp
	ytregex = regexp.MustCompile("(((www.)?youtube.com/.*(.v=|&v=).+)|(youtu.be/.+))")
	titleregex = regexp.MustCompile("<meta property=\"og:title\" content=\".*\">")

	// Check for duplicate entries
	iter := mapfs.MakeMapIter(quoteDb)
	for {
		k, v, ok := iter()
		if !ok {
			break
		}
		if quoteDb.CountMatches(v) > 1 {
			log.Fatal(fmt.Errorf("Duplicate entry in quote DB with key: %d - val: %s", k, v))
		}
	}

	iter = mapfs.MakeMapIter(urlDb)
	for {
		k, v, ok := iter()
		if !ok {
			break
		}
		if urlDb.CountMatches(v) > 1 {
			log.Fatal(fmt.Errorf("Duplicate entry in url DB with key: %d - val: %s", k, v))
		}
	}

	rand.Seed(time.Now().UnixNano())

	dispatch := map[string]func(irc.Message){
		botname:        cmdKunt,
		"!theo":	cmdTheoQuote,
		"!wisdom":      cmdWisdom,
		"!help":        cmdHelp,
		"!version":     cmdVersion,
		"!randquote":   cmdRandQuote,
		"!addquote":    cmdAddQuote,
		"!countquotes": cmdCountQuotes,
		"!randurl":     cmdRandUrl,
		"!addurl":      cmdAddUrl,
		"!counturls":   cmdCountUrls,
		"!radio":       cmdRadio,
		"!uptime":      cmdUptime,
		"!src":         cmdSrc,
		"!TODO":        cmdTodo,
		"!game":        cmdGame,
		"!copyright":   cmdCopyright,
	}
	if configfile == "" {
		hostport := strings.Split(flag.Arg(0), ":")

		cfg = &irc.Config{
			NetworkName: "freenode",
			Nick:        botname,
			User:        "z0mg",
			RealName:    botname,
			Serv:        hostport[0],
			Port:        hostport[1],
			Ssl:         *sslon,
			Channels:    []string{"#2f30"},
		}
	} else {
		cfg,err = irc.LoadConfig(configfile)
		if err != nil {
			log.Fatal(err)
		}
	}

	kunt.ircCtx = irc.NewIrcContext(*cfg)
	kunt.ircCtx.AddEventHandler(irc.Event{Command: "PRIVMSG", Fn: func(msg irc.Message) {
		cmd := msg.Args[1]
		for i, v := range dispatch {
			if strings.HasPrefix(cmd, i) {
				go v(msg)
			}
		}
	}})

	kunt.ircCtx.AddEventHandler(irc.Event{Command: "PING", Fn: func(msg irc.Message) {
		kunt.ircCtx.Pong(msg.Args[0])
	}})

	kunt.ircCtx.AddEventHandler(irc.Event{Command: "ERROR", Fn: func(msg irc.Message) {
		os.Exit(0)
	}})

	kunt.ircCtx.AddEventHandler(irc.Event{Command: "KICK", Fn: func(msg irc.Message) {
		time.Sleep(750 * time.Millisecond)
		kunt.ircCtx.Join("#2f30", "")
	}})

	err = kunt.ircCtx.Connect()
	if err != nil {
		log.Fatal(err)
	}
	kunt.ircCtx.Login()

	syscall.Mkfifo(".kfifo", 0600)
	go fifoCmdParse()

	donechan = make(chan bool)
	select {
	case stat := <-donechan:
		if stat {
			break
		}
	}

	kunt.ircCtx.Quit("releasing entropy")
	log.Println(" [Exiting] ")
}
